var Fd = {
  IMAGE_BLANK: FdFile.getAsDataURL("chrome://fastdial/skin/blank.gif"),
  parent: FdStorage.getRoot(),
  currentIndex: 0,
  prefsWindow: null,
  forceDrawPage: false,
  pageHtml: null,
  loading: new Array(),
  refreshTimeout: 0,

  initialize: function() {
    var chromehidden = $("main-window").getAttribute("chromehidden");
    if (chromehidden.match(/location/)) return; // this is a popup window

    Fd.initShortcutKeys();

    window.addEventListener("resize", Fd.updatePage, false);
    gBrowser.addEventListener("load", Fd.onPageLoad, true);
    $("contentAreaContextMenu").addEventListener("popupshowing",
                                                 Fd.onMainMenuShowing, false);
    var tabMenu = document.getAnonymousElementByAttribute(gBrowser, "anonid", "tabContextMenu");
    tabMenu.addEventListener("popupshowing", Fd.onTabMenuShowing, false);
    // FF2 - bookmarks, FF3 - places
    var bookmarksMenu = $("bookmarks-context-menu") || $("placesContext");
    bookmarksMenu.addEventListener("popupshowing", Fd.onBookmarksMenuShowing, false);

    if (FdPrefs.getIntPref("refresh") > 0) Fd.refreshAll();

    Fd.drawPagesOnStartup();
    Fd.showFF2notice();
  },

  refreshAll: function() {
    var pageInfos = FdStorage.getPageInfos();
    for(var i in pageInfos) {
      if (pageInfos[i].refresh) {
        Fd.refreshThumbnail(null, pageInfos[i]);
      }
    }
    Fd.updateRefreshTimeout();
  },

  updateRefreshTimeout: function() {
    if (Fd.refreshTimeout) {
      clearTimeout(Fd.refreshTimeout);
    }
    var time = FdPrefs.getIntPref("refresh");
    if (time > 0) Fd.refreshTimeout = setTimeout(Fd.refreshAll, time * 60 * 1000);
  },

  drawPagesOnStartup: function() {
    FdUtils.forEachBlankPage(function(doc) {
      var tab = FdUtils.getDocumentTab(doc);
      if (!tab.hasAttribute("busy") && FdUtils.isBlank(doc)) {
        Fd.drawPage(doc);
      }
    });
  },

  showFF2notice: function() {
    if (FdUtils.isFirefox2() &&
        FdPrefs.getCharPref("scaleMethod") == "default" &&
        FdPrefs.getBoolPref("showFF2notice"))
    {
      FdPrefs.setBoolPref("showFF2notice", false);
      openDialog("chrome://fastdial/content/ff2notice.xul",
               "fastdial-ff2notice-window",
               "chrome,centerscreen=yes");
    }
  },

  initShortcutKeys: function() {
    var modifiers = FdPrefs.getCharPref("shortcutKeys");
    for(var i = 1; i <= 9; i++) {
      var key = $("fastdial-key-" + i);
      key.setAttribute("modifiers", modifiers);
    }
  },

  open: function(event) {
    if (event.button == 0) {
      Fd.forceDrawPage = true;
      FdUtils.openUILink("about:blank");
    }
    else if (event.button == 1) {
      Fd.forceDrawPage = true;
      FdUtils.openUILink("about:blank", "tab");
    }
  },

  onPageLoad: function(event) {
    var doc = event.originalTarget;
    if (!FdUtils.getDocumentTab(doc) || // this is an iframe
        !FdUtils.isBlank(doc)) return;

    if (Fd.forceDrawPage ||
        FdPrefs.getBoolPref("showInNewTabs"))
    {
      Fd.forceDrawPage = false;
      Fd.drawPage(doc);
    }
  },

  drawPage: function(doc) {
    if (!Fd.pageHtml) {
      Fd.pageHtml = Fd.createPageHtml();
    }
    doc.body.innerHTML = Fd.pageHtml;
    Fd.updateDoc(doc);

    doc.addEventListener("click", Fd.onPageClick, false);
    doc.addEventListener("mousemove", FdPage.onMouseMove, false);
    doc.addEventListener("mousedown", FdPage.onMouseDown, false);
    doc.addEventListener("mouseup", FdPage.onMouseUp, false);
    doc.addEventListener("contextmenu", Fd.onMenuShowing, false);

    var title = FdBundle.getString("fastDial");
    FdURL.addToHistory("about:blank", title);

    var showIconAndTitle = FdPrefs.getBoolPref("showIconAndTitle");
    doc.title = showIconAndTitle ? title : "";
    setTimeout(function() {
      var tab = FdUtils.getDocumentTab(doc);
      tab.setAttribute("image", showIconAndTitle
                                  ? "chrome://fastdial/skin/fastdial_icon.png" : "");
    }, 0);
  },

  invalidatePageHtml: function() {
    Fd.pageHtml = null;
  },

  isBackButton: function(index, parent) {
    if (!parent) parent = Fd.parent;
    return parent.id && index == 1;
  },

  isVisible: function(pageInfo) {
    return Fd.isBackButton(pageInfo.index) ||
           !pageInfo.isEmpty() ||
           !FdPrefs.getBoolPref("hideEmpty");
  },

  getClass: function(pageInfo) {
    return Fd.isBackButton(pageInfo.index) ? "back" :
           pageInfo.isParent ? "parent" : "image";
  },

  createPageHtml: function() {
    Fd.parent = Fd.parent ? FdStorage.getPageInfo(null, null, Fd.parent.id)
                          : FdStorage.getRoot();
    var html = "";
    var hide = " style='display: none;'";

    for(var i = 1; i <= Fd.parent.width * Fd.parent.height; i++) {
      var pageInfo = Fd.parent.getChild(i);
      if (i % Fd.parent.width == 1) html += "<tr>";

      html += "<td>" +
              "<div class='cell' id='cell" + i + "'" + 
              (Fd.isVisible(pageInfo) ? "" : " style='visibility: hidden;'") + ">" +
              "<div class='buttons' id='buttons" + i + "'>" +
              "<div class='button clear' title='" +
              FdBundle.getString("clear") + "'></div>" +
              "<div class='button refresh' title='" +
              FdBundle.getString("refresh") + "'></div>" +
              "<div class='button properties' title='" +
              FdBundle.getString("properties") + "'></div>" +
              "</div>" +
              "<div class='loading' id='loading" + i + "'" +
              (Fd.loading[pageInfo.id] ? "" : hide) + "></div>" +
              "<a href='" + pageInfo.url + "' id='link" + i + "'" +
              (Fd.loading[pageInfo.id] ? hide : "") + ">" +
              "<img class='" + Fd.getClass(pageInfo) + "' id='" + i + "' src='" +
              (pageInfo.getImageURL() || Fd.IMAGE_BLANK) + "'></a>" +
              "<div class='title' id='title" + i + "'>" +
              (pageInfo.title || "&nbsp;") + "</div>" +
              "</div></td>";

      if (i % Fd.parent.width == 0) html += "</tr>";
    }
    return FdTheme.current.template.replace(/%content%/, html);
  },

  onPageClick: function(event) {
    var panel = new FdPanel(event.target);
    if (!panel.id || event.button == 2) return;

    event.stopPropagation();
    event.preventDefault();

    switch(event.target.className) {
      case "button clear":      return Fd.clearURL(panel.id);
      case "button refresh":    return Fd.refreshThumbnail(panel.id);
      case "button properties": return Fd.showProperties(panel.id);

      case "image":
      case "parent":
          var pageInfo = Fd.parent.getChild(panel.id);
          if (!pageInfo.isEmpty()) {
            var where = FdUtils.whereToOpenLink(event);
            return Fd.openPage(panel.id, where);
          }
      case "cell":
      case "title":
      case "back":
          if (Fd.isBackButton(panel.id)) return Fd.openPage(panel.id);
          if (event.button == 0) return Fd.showProperties(panel.id);
          if (event.button == 1) return Fd.refreshThumbnail(panel.id);
    }
  },

  movePageInfo: function(sourceIndex, targetIndex, moveInside) {
    var source = Fd.parent.getChild(sourceIndex);
    var target = Fd.parent.getChild(targetIndex);

    if (target.isParent && moveInside) {
      source.parentId = target.id;
      source.index = Fd.getEmptyPageInfo(target).index;
    }
    else if (Fd.isBackButton(target.index)) {
      source.parentId = Fd.parent.parentId;
      source.index = Fd.getEmptyPageInfo(Fd.parent.getParent()).index;
    }
    else {
      source.index = targetIndex;
      target.index = sourceIndex;
    }
    source.save();
    target.save();
    Fd.updateImage(sourceIndex);
    Fd.updateImage(targetIndex);
  },

  onMenuShowing: function(event) {
    Fd.currentIndex = FdPanel.getId(event.target);
    var pageInfo = Fd.currentIndex && Fd.parent.getChild(Fd.currentIndex);
    var isEmpty = !pageInfo || pageInfo.isEmpty();

    $("fastdial-menu-properties").hidden = !Fd.currentIndex ||
                                           Fd.isBackButton(Fd.currentIndex);
    $("fastdial-menu-open-tab").hidden = 
    $("fastdial-menu-open-tabshifted").hidden =
    $("fastdial-menu-open-separator").hidden = isEmpty || pageInfo.isParent;
    $("fastdial-menu-refresh").hidden =
    $("fastdial-menu-clear").hidden = isEmpty;
    $("fastdial-menu-share-logo").hidden = !pageInfo ||
                                           !pageInfo.customImage ||
                                           pageInfo.customImage.match(/^http:\/\/userlogos.org/);

    // Workaround for FF 2.x bug: otherwise popup is displayed in wrong place
    document.popupNode = null;

    var menu = $("fastdial-menu");
    menu.showPopup(menu, event.screenX, event.screenY, "context", null, null);

    event.stopPropagation();
    event.preventDefault();
  },

  onMainMenuShowing: function(event) {
    if (event.target.id == "contentAreaContextMenu") {
      var menu = $("fastdial-add-menu");
      menu.hidden = !FdPrefs.getBoolPref("showAddMenu");
      var url = gContextMenu.target.href || "";
      menu.setAttribute("oncommand", "Fd.addPage('" + url + "');");
    }
  },

  pageInfo: null,

  showProperties: function(index, pageInfo) {
    var oldPageInfo = Fd.parent.getChild(index);
    if (!Fd.pageInfo) Fd.pageInfo = oldPageInfo.clone();

    openDialog("chrome://fastdial/content/properties.xul",
               "fastdial-properties-window",
               "chrome,centerscreen=yes,modal=yes,resizable=yes");

    if (Fd.pageInfo) {
      if (!Fd.pageInfo.isEmpty()) {
        Fd.pageInfo.save();

        var doc = content.document;
        if (Fd.pageInfo.url != oldPageInfo.url || !Fd.pageInfo.title ||
            Fd.pageInfo.customImage != oldPageInfo.customImage ||
            Fd.pageInfo.customBackground != oldPageInfo.customBackground)
        {
          Fd.refreshThumbnail(index);
        }
        else if (Fd.pageInfo.title != oldPageInfo.title) {
          Fd.updateImage(index);
        }
      } else Fd.clearURL(index, true);
    }
    Fd.pageInfo = null;
  },

  refreshThumbnail: function(index, pageInfo) {
    if (!pageInfo) pageInfo = Fd.parent.getChild(index);
    if (pageInfo.isEmpty()) return;

    Fd.loading[pageInfo.id] = true;
    Fd.updateImage(pageInfo.index);

    if (pageInfo.customImage && pageInfo.title) {
      Fd.saveCustomImage(pageInfo);
    }
    else Fd.saveImage(pageInfo);
  },

  createFrame: function() {
    var browser = document.createElement("browser");
    browser.width = 1024;
    browser.height = 768;
    browser.setAttribute("type", "content");
    $("hidden-box").appendChild(browser);
    return browser;
  },

  deleteFrame: function(iframe) {
    iframe.parentNode.removeChild(iframe);
  },

  loadInFrame: function(url, onLoad) {
    function onFrameLoad() {
      iframe.removeEventListener("load", onFrameLoad, true);
      clearTimeout(loadTimeout);
      onLoad(iframe);
    }
    var iframe = Fd.createFrame();
    iframe.addEventListener("load", onFrameLoad, true);

    var loadTimeout = setTimeout(onFrameLoad,
                                 FdPrefs.getIntPref("maxLoadTime") * 1000);
    try {
      iframe.setAttribute("src", url);
    }
    catch(e) {
      FdUtils.error("incorrectURL", [url]);
      onFrameLoad();
    }
  },

  saveImage: function(pageInfo) {
    Fd.loadInFrame(pageInfo.url, function(iframe) {
        if (!pageInfo.title) {
          var doc = iframe.contentDocument;
          pageInfo.title = doc.title || doc.location.href;
          pageInfo.save();
        }
        if (pageInfo.customImage) {
          Fd.deleteFrame(iframe);
          Fd.saveCustomImage(pageInfo);
        }
        else Fd.saveThumbnail(pageInfo, iframe);
      });
  },

  saveCustomImage: function(pageInfo) {
    var url = pageInfo.customImage + "#" + new Date().getTime(); // Workaround cache

    Fd.loadInFrame(url, function(iframe) {
        var doc = iframe.contentDocument;
        var img = doc.body.firstChild;
        if (FdThumbnail.getHeight(img.width) / img.height > 1) {
          doc.body.style.width = img.width;
          doc.body.style.height = FdThumbnail.getHeight(img.width);
        }
        else {
          doc.body.style.width = FdThumbnail.getWidth(img.height);
          doc.body.style.height = img.height;
        }
        doc.body.style.background = pageInfo.customBackground || "white";
        doc.body.style.margin = 0;
        doc.body.style.display = "table-cell";
        doc.body.style.textAlign = "center";
        doc.body.style.verticalAlign = "middle";

        Fd.saveThumbnail(pageInfo, iframe);
      });
  },

  saveThumbnail: function(pageInfo, iframe) {
    var renderTimeout = FdPrefs.getIntPref("maxLoadTime") * 1000 / 5;

    setTimeout(function() {
        FdThumbnail.createThumbnail(iframe.contentWindow, Fd.parent.thumbWidth,
          function(result) {
            Fd.deleteFrame(iframe);
            pageInfo.saveBigImage(result && result.big);
            pageInfo.saveImage(result && result.image);
            delete Fd.loading[pageInfo.id];
            Fd.updateImage(pageInfo.index);
          });
      },
      renderTimeout);
  },

  updateImage: function(index) {
    FdUtils.forEachBlankPage(function(doc) {
      var image = doc.getElementById(index);
      if (image) {
        var pageInfo = Fd.parent.getChild(index);
        image.className = Fd.getClass(pageInfo);

        var url = pageInfo.getImageURL();
        FdURL.removeFromCache(url);
        image.src = url || Fd.IMAGE_BLANK;

        var panel = new FdPanel(image);
        var link = panel.get("link");
        link.href = pageInfo.url;
        link.style.display = Fd.loading[pageInfo.id] ? "none" : "block";

        panel.get("loading").style.display = Fd.loading[pageInfo.id] ? "block" : "none";
        panel.get("title").innerHTML = pageInfo.title || "&nbsp;";
      }
    });
    Fd.invalidatePageHtml();
  },

  clearURL: function(index, dontAskConfirmation) {
    var pageInfo = Fd.parent.getChild(index);
    if (pageInfo.isEmpty()) return;

    var title = pageInfo.title || pageInfo.url;
    var message = FdBundle.getString("clearConfirmation", [title]);
    if (dontAskConfirmation || FdUtils.confirm(message)) {
      FdStorage.deletePageInfo(pageInfo);
      Fd.updateImage(pageInfo.index);
    }
  },

  getImageSize: function(doc) {
    var width, height;

    var body  = FdUtils.getComputedStyle(doc.body),
        cntnt = FdUtils.getComputedStyle(doc.getElementById("content")),
        image = FdUtils.getComputedStyle(doc.getElementById("1")),
        cell  = FdUtils.getComputedStyle(doc.getElementById("cell1")),
        title = FdUtils.getComputedStyle(doc.getElementById("title1"));

    if (Fd.parent.fixedSize) {
      width = Fd.parent.thumbWidth;
      height = FdThumbnail.getHeight(width);
    }
    else {
      width = (content.innerWidth - body.getInt("margin-left")*2 -
              cntnt.getInt("border-spacing")) / Fd.parent.width -
              cell.getInt("border-left-width")*2 - cell.getInt("padding-left")*2 -
              image.getInt("border-left-width")*2 - cntnt.getInt("border-spacing");

      height = (content.innerHeight - body.getInt("margin-top")*2 -
               cntnt.getInt("border-spacing")) / Fd.parent.height -
               cell.getInt("border-left-width")*2 - cell.getInt("padding-top") -
               image.getInt("border-left-width")*2 - cntnt.getInt("border-spacing") -
               title.getInt("margin-top")*2 - title.getInt("height");
    }
    width = Math.max(Math.floor(width), FdThumbnail.MIN_WIDTH);
    height = Math.max(Math.floor(height), FdThumbnail.MIN_HEIGHT);

    if (FdThumbnail.getHeight(width) / height > 1) {
      width = FdThumbnail.getWidth(height);
    }
    else height = FdThumbnail.getHeight(width);

    return {
      image: { width: width, height: height },
      title: { width: width + image.getInt("border-left-width")*2 -
                              title.getInt("margin-left")*2 }
    }
  },

  updatePage: function() {
    FdUtils.forEachBlankPage(Fd.updateDoc);
  },

  redrawPage: function() {
    Fd.invalidatePageHtml();
    FdUtils.forEachBlankPage(Fd.drawPage);
  },

  updateDoc: function(doc) {
    for(var i = 0; i < doc.styleSheets.length; i++) {
      var styleSheet = doc.styleSheets[i];
      if (styleSheet.href &&
          styleSheet.href != "about:blank") continue;

      var rules = styleSheet.cssRules;
      for(var j = 0; j < rules.length; j++) {
        Fd.applyStyle(rules[j], FdTheme.current.style);
      }
        
      var size = Fd.getImageSize(doc);

      for(var j = 0; j < rules.length; j++) {
        var rule = rules[j];
        switch(rule.selectorText) {
          case ".image, .loading, .back, .parent":
              rule.style.height = size.image.height;
              rule.style.width = size.image.width;
              break;
          case ".title":
              rule.style.width = size.title.width;
              break;
        }
      }
    }
  },

  applyStyle: function(rule, style) {
    if (!rule.selectorText) return;

    var regexp = /\s*([^#.]+)\s*(,|$)/g;
    while(match = regexp.exec(rule.selectorText)) {
      var entry = style[match[1]];
      if (!entry) continue;

      if (rule.selectorText.match(/hover/)) {
        entry = entry.hover;
      }
      for(var i in entry) {
        rule.style[i] = entry[i];
      }
    }
  },

  openPreferences: function() {
    if (!Fd.prefsWindow || Fd.prefsWindow.closed) {
      Fd.prefsWindow = window.openDialog("chrome://fastdial/content/preferences.xul",
                                         "fastdial-preferences-window",
                                         "chrome,toolbar,centerscreen,resizable=yes");
    } else Fd.prefsWindow.focus();
  },

  openPage: function(index, where) {
    var pageInfo = Fd.parent.getChild(index);
    if (pageInfo.isParent) {
      Fd.parent = pageInfo;
      Fd.redrawPage();
    }
    else if (Fd.isBackButton(index)) {
      Fd.parent = Fd.parent.getParent();
      Fd.redrawPage();
    }
    else {
      if (!where) where = FdPrefs.getCharPref("openIn");
      FdUtils.openUILink(pageInfo.url, where);
    }
  },

  addPage: function(url, title) {
    var pageInfo = Fd.getEmptyPageInfo(Fd.parent);
    pageInfo.url = url || content.document.location.href;
    pageInfo.title = url ? title : content.document.title;
    pageInfo.save();
    Fd.refreshThumbnail(null, pageInfo);
  },

  findLogo: function() {
    var url = "http://userlogos.org/";
    if (Fd.currentIndex) {
      var pageInfo = Fd.parent.getChild(Fd.currentIndex);
      if (pageInfo.url) {
        var host = FdURL.getNsiURL(pageInfo.url).host;
        url = "http://userlogos.org/search/node/" + encodeURIComponent(host + " type:logo");
      }
    }
    FdUtils.openUILink(url, "tab");
  },

  shareLogo: function() {
    FdUtils.openUILink("http://userlogos.org/node/add/logo", "tab");
  },

  showZoom: function(index) {
    var pageInfo = Fd.parent.getChild(index);
    if (pageInfo.isEmpty()) return;

    var doc = content.document;
    var zoom = doc.createElement("div");
    zoom.id = "zoom";
    zoom.innerHTML = "<div id='zoom-background'></div>" + 
                     "<div id='zoom-image'>" +
                     "<img src='" + pageInfo.getBigImageURL() + "'></div>";
    doc.body.appendChild(zoom);

    function onMouseUp(event) {
      var zoom = doc.getElementById("zoom");
      zoom.parentNode.removeChild(zoom);
      doc.removeEventListener("mouseup", onMouseUp, true);
      event.preventDefault();
    }
    doc.addEventListener("mouseup", onMouseUp, true);
  },

  onBookmarksMenuShowing: function(event) {
    var menuitem = $("fastdial-add-bookmark-menu");
    // Add menu item manually for FF2
    if (!menuitem) { 
      menuitem = $("fastdial-add-menu").cloneNode(true);
      menuitem.setAttribute("id", "fastdial-add-bookmark-menu");
      menuitem.setAttribute("oncommand", "Fd.addBookmark();");
      event.target.appendChild(menuitem);
    }
  },

  addBookmark: function() {
    function add(bookmark) {
      var url = bookmark.getAttribute("statustext");
      if (url) Fd.addPage(url, bookmark.getAttribute("label"));
    }
    var bookmark = document.popupNode;
    if (!bookmark.getAttribute("container")) add(bookmark);
    else {
      var menupopup = bookmark.firstChild;
      for(var i = 0; i < menupopup.childNodes.length; i++) {
        add(menupopup.childNodes[i]);
      }
    }
  },

  getEmptyPageInfo: function(parent) {
    for(var i = 1;; i++) {
      var pageInfo = parent.getChild(i);
      if (!Fd.isBackButton(i, parent) && pageInfo.isEmpty()) {
        if (i > parent.width * parent.height) {
          parent.height = Math.ceil(i / parent.width);
          parent.save();
          Fd.redrawPage();
        }
        return pageInfo;
      }
    }
  },

  onTabMenuShowing: function(event) {
    var menuitem = $("fastdial-add-tab-menu");
    if (!menuitem) {
      menuitem = $("fastdial-add-menu").cloneNode(true);
      menuitem.setAttribute("id", "fastdial-add-tab-menu");
      menuitem.setAttribute("oncommand", "Fd.addTab();");
      event.target.appendChild(menuitem);
    }
  },

  addTab: function() {
    var tab = document.popupNode;
    var doc = tab.linkedBrowser.contentDocument;
    Fd.addPage(doc.location.href, doc.title);
  }
}
