var FdPrefs = {
  prefs: Components.classes["@mozilla.org/preferences-service;1"]
         .getService(Components.interfaces.nsIPrefService)
         .getBranch("extensions.fastdial."),

  clearPageInfo: function(id) {
    this.clearPref(id);
  },

  clearPref: function(name) {
    try {
      this.prefs.deleteBranch(name);
    } catch(e) {}
  },

  getBoolPref: function(name) {
    try {
      return this.prefs.getBoolPref(name);
    } catch(e) {
      return false;
    }
  },

  setBoolPref: function(name, value) {
    this.prefs.setBoolPref(name, value);
  },

  getIntPref: function(name) {
    try {
      return this.prefs.getIntPref(name);
    } catch(e) {
      return 0;
    }
  },

  setIntPref: function(name, value) {
    this.prefs.setIntPref(name, value);
  },

  getCharPref: function(name) {
    try {
      return this.prefs.getComplexValue(name,
        Components.interfaces.nsISupportsString).data;
    } catch(e) {
      return null;
    }
  },

  setCharPref: function(name, value) {
    this.clearPref(name); // Remove legacy preference

    var str = Components.classes["@mozilla.org/supports-string;1"]
              .createInstance(Components.interfaces.nsISupportsString);
    str.data = value;
    this.prefs.setComplexValue(name,
      Components.interfaces.nsISupportsString, str);
  },

  setObjectPref: function(name, object) {
    this.setCharPref(name, FdUtils.toJSON(object));
  },

  getObjectPref: function(name) {
    return eval(this.getCharPref(name));
  }
}


var FdFile = {
  getDataDirectory: function() {
    var file = Components.classes["@mozilla.org/file/directory_service;1"]
               .getService(Components.interfaces.nsIProperties)
               .get("ProfD", Components.interfaces.nsIFile);
    file.append("fastdial");
    if (!file.exists()) {
      file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0777);
    }
    return file;
  },

  getURL: function(name) {
    return "chrome://fastdial-profile/content/" + encodeURIComponent(name);
  },

  getExtensionDirectory: function() {
    var id = "fastdial@telega.phpnet.us";
    return Components.classes["@mozilla.org/extensions/manager;1"]
           .getService(Components.interfaces.nsIExtensionManager)
           .getInstallLocation(id)
           .getItemLocation(id);
  },

  getThemesDirectory: function() {
    var file = FdFile.getExtensionDirectory();
    file.append("chrome");
    file.append("content");
    file.append("themes");
    return file;
  },

  writeFile: function(file, data, unichar) {
    var out = Components.classes["@mozilla.org/network/file-output-stream;1"]
              .createInstance(Components.interfaces.nsIFileOutputStream);
    out.init(file, 0x04 | 0x08 | 0x20, 0666, 0); // read & write, create, truncate

    if (unichar) {
      var uniOut = Components.classes["@mozilla.org/intl/converter-output-stream;1"]
                    .createInstance(Components.interfaces.nsIConverterOutputStream);
      uniOut.init(out, "utf-8", 0, 0xFFFD);
      uniOut.writeString(data);
      uniOut.close();
    }
    else {
      out.write(data, data.length);
    }
    out.close();
  },

  getNsiFile: function(file) {
    if (file instanceof Components.interfaces.nsIFile) return file;
    else {
      var nsiFile = Components.classes["@mozilla.org/file/local;1"]
                .createInstance(Components.interfaces.nsILocalFile);
      nsiFile.initWithPath(file);
      return nsiFile;
    }
  },

  getFileURL: function(file) {
    var nsiFile = FdFile.getNsiFile(file);
    var ios = Components.classes["@mozilla.org/network/io-service;1"]
              .getService(Components.interfaces.nsIIOService);
    var fileHandler = ios.getProtocolHandler("file")
                     .QueryInterface(Components.interfaces.nsIFileProtocolHandler);
    return fileHandler.getURLSpecFromFile(nsiFile);
  },

  read: function(urlOrFile, unichar) {
    var url = (urlOrFile instanceof Components.interfaces.nsIFile)
                ? FdFile.getFileURL(urlOrFile)
                : urlOrFile;

    var ioService = Components.classes["@mozilla.org/network/io-service;1"]
                    .getService(Components.interfaces.nsIIOService);
    var channel = ioService.newChannel(url, null, null);
    var stream = channel.open();

    function readUnicharStream(stream) {
      var unichar = Components.classes["@mozilla.org/intl/converter-input-stream;1"]
                    .createInstance(Components.interfaces.nsIConverterInputStream);
      unichar.init(stream, "utf-8", stream.available(), 0xFFFD);
      var str = {};
      unichar.readString(stream.available(), str);
      unichar.close();
      return str.value;
    }

    function readBinaryStream(stream) {
      var binary = Components.classes["@mozilla.org/binaryinputstream;1"]
                   .createInstance(Components.interfaces.nsIBinaryInputStream);
      binary.setInputStream(stream);
      var bytes = binary.readBytes(binary.available());
      binary.close();
      return bytes;
    }

    var data = unichar ? readUnicharStream(stream)
                       : readBinaryStream(stream);
    stream.close();
    return data;
  },

  getAsDataURL: function(urlOrFile) {
    var data = FdFile.read(urlOrFile);
    return "data:image/png;base64," + btoa(data);
  },

  chooseFile: function(mode, filters, name) {
    var fp = Components.classes["@mozilla.org/filepicker;1"]
             .createInstance(Components.interfaces.nsIFilePicker);
    fp.init(window, null, mode == "save" ? fp.modeSave :
                          mode == "folder" ? fp.modeGetFolder : fp.modeOpen);
    for(var i in filters) {
      switch(filters[i]) {
        case "images": fp.appendFilters(fp.filterImages); break;
        case "html":   fp.appendFilters(fp.filterHTML); break;
        default:       fp.appendFilter(filter, filter); break;
      }
    }
    fp.appendFilters(fp.filterAll);
    fp.defaultString = name;

    var result = fp.show();
    if (result == fp.returnOK ||
        result == fp.returnReplace) return fp.file;
  },

  forEachFile: function(dir, onFile) {
    var files = dir.directoryEntries;
    while(files.hasMoreElements()) {
      var file = files.getNext();
      file.QueryInterface(Components.interfaces.nsIFile);
      onFile(file);
    }
  }
}

var FdURL = {
  getNsiURL: function(url) {
    var nsiUrl = Components.classes["@mozilla.org/network/standard-url;1"]
                 .createInstance(Components.interfaces.nsIURL);
    nsiUrl.spec = url;
    return nsiUrl;
  },

  getFileName: function(url) {
    var nsiUrl = (nsiUrl instanceof Components.interfaces.nsIURL)
                   ? url : FdURL.getNsiURL(url);
    return decodeURIComponent(nsiUrl.fileName);
  },

  getFaviconURL: function(url) {
    if (url) {
      var nsiUrl = FdURL.getNsiURL(url);
      if (nsiUrl.scheme != "ftp") {
        return nsiUrl.prePath + "/favicon.ico";
      }
    }
  },

  getScheme: function(url) {
    if (url) {
      return FdURL.getNsiURL(url).scheme;
    }
  },

  isImage: function(url) {
    return url && url.match(/\.(jpg|jpeg|png|gif|bmp)$/i);
  },

  removeFromCache: function(url) {
    if (!url) return;

    var clientId = url.match(/^chrome:/) ? "image-chrome" : "image";
    var cacheSession = Components.classes["@mozilla.org/network/cache-service;1"]
                       .getService(Components.interfaces.nsICacheService)
                       .createSession(clientId, 0, false);
    try {
      var entry = cacheSession.openCacheEntry(url,
                    Components.interfaces.nsICache.ACCESS_READ, false);
      entry.doom();
    }
    catch(e) {}
  },

  addToHistory: function(url, title) {
    var sessionHistory = gBrowser.sessionHistory;
    sessionHistory.QueryInterface(Components.interfaces.nsISHistoryInternal);

    // nsIURL doesn't work for "about:blank" url
    var nsiUri = Components.classes["@mozilla.org/network/simple-uri;1"]
                 .createInstance(Components.interfaces.nsIURI);
    nsiUri.spec = url;
    var entry = Components.classes["@mozilla.org/browser/session-history-entry;1"]
                .createInstance(Components.interfaces.nsISHEntry);
    entry.setURI(nsiUri);
    entry.setTitle(title);

    if (!sessionHistory.count) {
      sessionHistory.addEntry(entry, true);
    } else {
      sessionHistory.replaceEntry(sessionHistory.index, entry);
    }
  }
}
