function FdPageInfo(id, parentId, index, url, title, refresh,
                    customImage, customBackground, isParent,
                    width, height, thumbWidth, fixedSize)
{
  var self = this;

  this.id = id;
  this.parentId = parentId;
  this.index = index;
  this.url = url || "";
  this.title = title || "";
  this.refresh = refresh;
  this.customImage = customImage || "";
  this.customBackground = customBackground || "";
  this.isParent = isParent;
  this.width = width;
  this.height = height;
  this.thumbWidth = thumbWidth;
  this.fixedSize = fixedSize;

  this.clone = function() {
    return new FdPageInfo(self.id, self.parentId, self.index, self.url, self.title,
                          self.refresh, self.customImage, self.customBackground,
                          self.isParent, self.width, self.height, self.thumbWidth,
                          self.fixedSize);
  }

  this.getParent = function() {
    return FdStorage.getPageInfo(null, null, self.parentId);
  }

  this.getChild = function(index) {
    return FdStorage.getPageInfo(self.id, index);
  }

  this.save = function() {
    FdStorage.savePageInfo(self);
  }

  this.isEmpty = function() {
    return !self.url && !self.isParent;
  }

  function getImageName(big) {
    return self.id + (big ? "_big" : "") + ".png";
  }

  function getImageFile(big) {
    var file = FdFile.getDataDirectory();
    file.append(getImageName(big));
    return file;
  }

  this.deleteImage = function(big) {
    var file = getImageFile(big);
    try {
      file.remove(false);
    }
    catch(e) {}
  }

  // "chrome://" images in FF 2.x expire from cache after 1 minute,
  // so use "data:" urls to avoid expiration

  this.getImageURL = function(big) {
    if (self.isEmpty()) return;
    var url = FdFile.getURL(getImageName(big));
    try {
      return FdUtils.isFirefox2() ? FdFile.getAsDataURL(url) : url;
    }
    catch(e) {}
  }

  this.saveImage = function(data, big) {
    var file = getImageFile(big);
    FdFile.writeFile(file, data || "");
  }

  this.getBigImageURL = function() {
    return this.getImageURL(true);
  }

  this.deleteBigImage = function() {
    this.deleteImage(true);
  }

  this.saveBigImage = function(data) {
    this.saveImage(data, true);
  }
}

function FdStorageImpl() {
  var self = this;

  var file = FdFile.getDataDirectory();
  file.append("fastdial.sqlite");
  this.firstRun = !file.exists();

  var storageService = Components.classes["@mozilla.org/storage/service;1"]
                       .getService(Components.interfaces.mozIStorageService);
  var connection = storageService.openDatabase(file);

  if (this.firstRun) {
    connection.executeSimpleSQL("CREATE TABLE pageInfos (" +
                                  "id               INTEGER PRIMARY KEY AUTOINCREMENT," +
                                  "parentId         INTEGER," +
                                  "indexx           INTEGER," +
                                  "url              STRING," +
                                  "title            STRING," +
                                  "refresh          INTEGER," +
                                  "customImage      STRING," +
                                  "customBackground STRING," +
                                  "isParent         BOOLEAN," +
                                  "width            INTEGER," +
                                  "height           INTEGER," +
                                  "thumbWidth       INTEGER," +
                                  "fixedSize        BOOLEAN)");
  }

  this.migrateLegacyPrefs = function() {
    var count = FdPrefs.getIntPref("width") * FdPrefs.getIntPref("height");
    for(var i = 1; i <= count; i++) {
      var pageInfo = FdPrefs.getObjectPref(i);
      if (pageInfo) {
        pageInfo = new FdPageInfo(null, 0, i,
                                  decodeURI(pageInfo.url),
                                  pageInfo.title,
                                  pageInfo.refresh,
                                  decodeURI(pageInfo.customImage),
                                  pageInfo.customBackground);
        self.savePageInfo(pageInfo);

        if (pageInfo.id == i) continue;

        try {
          var file = FdFile.getDataDirectory();
          file.append(i + ".png");
          file.moveTo(null, pageInfo.id + ".png");
        }
        catch(e) {}
        try {
          var file = FdFile.getDataDirectory();
          file.append(i + "_big.png");
          file.moveTo(null, pageInfo.id + "_big.png");
        }
        catch(e) {}
      }
    }
  }

  try { // Since version 1.83
    connection.executeSimpleSQL("ALTER TABLE pageInfos ADD width INTEGER");
    connection.executeSimpleSQL("ALTER TABLE pageInfos ADD height INTEGER");
    connection.executeSimpleSQL("ALTER TABLE pageInfos ADD thumbWidth INTEGER");
    connection.executeSimpleSQL("ALTER TABLE pageInfos ADD fixedSize BOOLEAN");
  }
  catch(e) {}

  var text = "text: ";

  function bindString(statement, id, value) {
    // Force 'text' type for strings with leading zeroes, otherwise SQLite strips these zeroes
    statement.bindStringParameter(id, (value && value[0] == "0") ? text + value : value);
  }

  function getString(statement, id) {
    var value = statement.getString(id);
    return (value && value.substr(0, text.length) == text) ? value.substr(text.length) : value;
  }

  this.getPageInfos = function(parentId) {
    var sql = "SELECT id, parentId, indexx, url, title, refresh, customImage," +
              "customBackground, isParent, width, height, thumbWidth, fixedSize " +
              "FROM pageInfos" + (parentId ? " WHERE parentId=?1" : "");
    var statement = connection.createStatement(sql);
    try {
      var pageInfos = new Array();
      if (parentId) statement.bindInt32Parameter(0, parentId);
      while (statement.executeStep()) {
        pageInfos.push(new FdPageInfo(statement.getInt32(0),
                                      statement.getInt32(1),
                                      statement.getInt32(2),
                                      getString(statement, 3),
                                      getString(statement, 4),
                                      statement.getInt32(5),
                                      getString(statement, 6),
                                      getString(statement, 7),
                                      statement.getInt32(8),
                                      statement.getInt32(9),
                                      statement.getInt32(10),
                                      statement.getInt32(11),
                                      statement.getInt32(12)));
      }
      return pageInfos;
    }
    finally {
      statement.reset();
    }
  }

  this.getPageInfo = function(parentId, index, id) {
    var sql = "SELECT id, parentId, indexx, url, title, refresh, customImage," +
              "customBackground, isParent, width, height, thumbWidth, fixedSize " +
              "FROM pageInfos WHERE " + (id != null ? "id=?1" : "parentId=?1 AND indexx=?2");
    var statement = connection.createStatement(sql);
    try {
      if (id != null) {
        statement.bindInt32Parameter(0, id);
      } else {
        statement.bindInt32Parameter(0, parentId);
        statement.bindInt32Parameter(1, index);
      }
      if (statement.executeStep()) {
        return new FdPageInfo(statement.getInt32(0),
                              statement.getInt32(1),
                              statement.getInt32(2),
                              getString(statement, 3),
                              getString(statement, 4),
                              statement.getInt32(5),
                              getString(statement, 6),
                              getString(statement, 7),
                              statement.getInt32(8),
                              statement.getInt32(9),
                              statement.getInt32(10),
                              statement.getInt32(11),
                              statement.getInt32(12));
      }
      else return new FdPageInfo(null, parentId, index);
    }
    finally {
      statement.reset();
    }
  }

  this.savePageInfo = function(pageInfo) {
    if (pageInfo.isEmpty()) {
      return this.deletePageInfo(pageInfo);
    }
    if (!pageInfo.isParent) {
      deleteChildren(pageInfo.id);
    }
    var sql = pageInfo.id != null ? "UPDATE pageInfos SET parentId=?1, indexx=?2, url=?3," +
                                    "title=?4, refresh=?5, customImage=?6, customBackground=?7," +
                                    "isParent=?8, width=?9, height=?10, thumbWidth=?11," +
                                    "fixedSize=?12 WHERE id=?13"
                                  : "INSERT INTO pageInfos (parentId, indexx, url, title," +
                                    "refresh, customImage, customBackground, isParent, width," +
                                    "height, thumbWidth, fixedSize) VALUES " +
                                    "(?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12)";
    var statement = connection.createStatement(sql);
    statement.bindInt32Parameter(0, pageInfo.parentId);
    statement.bindInt32Parameter(1, pageInfo.index);
    bindString(statement, 2, pageInfo.url);
    bindString(statement, 3, pageInfo.title);
    statement.bindInt32Parameter(4, pageInfo.refresh);
    bindString(statement, 5, pageInfo.customImage);
    bindString(statement, 6, pageInfo.customBackground);
    statement.bindInt32Parameter(7, pageInfo.isParent);
    statement.bindInt32Parameter(8, pageInfo.width);
    statement.bindInt32Parameter(9, pageInfo.height);
    statement.bindInt32Parameter(10, pageInfo.thumbWidth);
    statement.bindInt32Parameter(11, pageInfo.fixedSize);
    if (pageInfo.id != null) statement.bindInt32Parameter(12, pageInfo.id);
    statement.execute();

    if (pageInfo.id == null) {
      var statement = connection.createStatement("SELECT last_insert_rowid() FROM pageInfos");
      try {
        if (statement.executeStep()) pageInfo.id = statement.getInt32(0);
      }
      finally {
        statement.reset();
      }
    }
  }

  function deleteChildren(parentId) {
    if (!parentId) return;

    var children = self.getPageInfos(parentId);
    for(var i in children) {
      self.deletePageInfo(children[i]);
    }
  }

  this.deletePageInfo = function(pageInfo) {
    if (!pageInfo.id) return;

    if (pageInfo.isParent) {
      deleteChildren(pageInfo.id);
    }
    var statement = connection.createStatement("DELETE FROM pageInfos WHERE id=?1");
    statement.bindInt32Parameter(0, pageInfo.id);
    statement.execute();

    pageInfo.deleteImage();
    pageInfo.deleteBigImage();
  }

  this.getRoot = function() {
    return this.getPageInfo(null, null, 0);
  }

  this.createRoot = function() {
    var root = new FdPageInfo();
    root.parentId = 0;
    root.isParent = true;
    this.savePageInfo(root);

    var statement = connection.createStatement("UPDATE pageInfos SET id=?1 WHERE id=?2");
    statement.bindInt32Parameter(0, 0);
    statement.bindInt32Parameter(1, root.id);
    statement.execute();

    var sql = "UPDATE pageInfos SET width=?1, height=?2, thumbWidth=?3, fixedSize=?4 " +
              "WHERE isParent = 1";
    var statement = connection.createStatement(sql);
    statement.bindInt32Parameter(0, FdPrefs.getIntPref("width"));
    statement.bindInt32Parameter(1, FdPrefs.getIntPref("height"));
    statement.bindInt32Parameter(2, FdPrefs.getIntPref("thumbWidth"));
    statement.bindInt32Parameter(3, FdPrefs.getBoolPref("fixedSize"));
    statement.execute();
  }
}

var FdStorage = new FdStorageImpl();
if (FdStorage.firstRun) FdStorage.migrateLegacyPrefs();
if (FdStorage.getRoot().isEmpty()) FdStorage.createRoot();
