function FdThemePref(pref) {
  this.id = pref && pref.id || "default";
  this.style = pref && pref.style || { body: {},
                                       content: {},
                                       cell: { hover: {} },
                                       image: { hover: {} }
                                     };
  removeLegacyProps(this.style);

  function removeLegacyProps(props) {
    if (typeof props != "object") return;
    for(var i in props) {
      if (i == "background") {
        props["backgroundColor"] = props[i];
        delete props[i];
      }
      else removeLegacyProps(props[i]);
    }
  }

  this.getValue = function(id) {
    var pref = this.style;
    var ids = id.split(/\./);
    for(var i = 0; i < ids.length; i++) {
      pref = pref[ids[i]];
    }
    var value = null;
    if (pref) {
      var match = pref.match(/\S+$/);
      if (match) value = match[0];
    }
    return value;
  }

  this.setValue = function(id, value) {
    var pref = this.style;
    var ids = id.split(/\./);
    for(var i = 0; i < ids.length - 1; i++) {
      pref = pref[ids[i]];
    }
    pref[ids[i]] = (value && ids[i] == "border")
                     ? "1px solid " + value : value;
    return value;
  }

  this.getImageName = function(id) {
    var name;
    var value = this.getValue(id);
    if (value) {    
      var url = value.match(/url\((.*?)\)/);
      if (url) name = FdURL.getFileName(url[1]);
    }
    return name || null;
  }

  this.setImageName = function(id, name) {
    var url = name ? "url(" + FdFile.getURL(name) + ")" : null;
    this.setValue(id, url);
  }
}

FdTheme = { 
  URL: "chrome://fastdial/content/themes/",
  _themes: null,
  _current: null,

  get themes() {
    if (!FdTheme._themes) {
      FdTheme._themes = FdTheme.loadThemes();
    }
    return FdTheme._themes;
  },

  get current() {
    if (!FdTheme._current) {
      var themePref = FdTheme.getPref();
      var theme = FdTheme.loadTheme(themePref.id);
      FdTheme._current = FdUtils.merge(theme, themePref);
    }
    return FdTheme._current;
  },

  getPref: function() {
    var pref = FdPrefs.getObjectPref("theme");
    return new FdThemePref(pref);
  },

  savePref: function(themePref) {
    themePref.style.buttons = themePref.style.cell;
    themePref.style.loading = themePref.style.image;
    themePref.style.back = themePref.style.image;

    FdPrefs.setObjectPref("theme", themePref);
    FdTheme._current = FdUtils.merge(FdTheme.themes[themePref.id], themePref);
  },

  loadTheme: function(id) {
    try {
      var template = FdFile.read(FdTheme.URL + id + "/template.html", true);

      return { id: id,
               template: template,
               name: FdBundle.getString(id + "Theme") || id
             };
    }
    catch(e) {
      FdUtils.error("themeError", [id, e]);
    }
  },

  loadThemes: function() {
    var themes = new Array();
    var themeDir = FdFile.getThemesDirectory();
    FdFile.forEachFile(themeDir, function(file) {
      if (file.isDirectory()) {
        var id = file.leafName;
        var theme = FdTheme.loadTheme(id);
        if (theme) themes[id] = theme;
      }
    });
    return themes;
  }
}
