var FdThumbnail = {
  RATIO: 0.75,
  MIN_WIDTH: 80,
  MIN_HEIGHT: 60,

  getHeight: function(width) {
    return Math.floor(width * FdThumbnail.RATIO);
  },

  getWidth: function(height) {
    return Math.floor(height / FdThumbnail.RATIO);
  },

  createBase64Thumb: function(wnd, thumbWidth) {
    if (thumbWidth == null) thumbWidth = FdThumbnail.getWindowWidth(wnd);

    var canvas = document.createElementNS("http://www.w3.org/1999/xhtml", "canvas");
    canvas.width = thumbWidth;
    canvas.height = FdThumbnail.getHeight(thumbWidth);

    var context = canvas.getContext("2d");
    context.clearRect(0, 0, canvas.width, canvas.height);

    var width = FdThumbnail.getWindowWidth(wnd);
    var height = FdThumbnail.getHeight(width);
    context.scale(canvas.width / width, canvas.height / height);
    context.drawWindow(wnd, 0, 0, width, height, "white");

    return canvas.toDataURL("image/png")
           .replace(/^data:image\/png;base64,/, "");
  },

  getWindowWidth: function(wnd) {
    var doc = wnd.document;
    var style = FdUtils.getComputedStyle(doc.body || doc.documentElement);
    return style.getInt("margin-left") +
           style.getInt("margin-right") +
           style.getInt("width");
  },

  createCanvasThumb: function(wnd, thumbWidth) {
    return { big: atob(FdThumbnail.createBase64Thumb(wnd)),
             image: atob(FdThumbnail.createBase64Thumb(wnd, thumbWidth)) };
  },

  // Create full image using javascript, then scale the image using java
  // Note: couldn't find a way to pass binary strings to/from Java code,
  // so passed base64 encoded strings instead

  createJavaThumb: function(wnd, thumbWidth) {
    var big = FdThumbnail.createBase64Thumb(wnd);
    var bytes = new sun.misc.BASE64Decoder().decodeBuffer(big);

    var byteIn = new java.io.ByteArrayInputStream(bytes);
    var image = Packages.javax.imageio.ImageIO.read(byteIn);

    var thumbHeight = FdThumbnail.getHeight(thumbWidth);
    var tempImage = image.getScaledInstance(thumbWidth, thumbHeight, image.SCALE_SMOOTH);
    var result = new java.awt.image.BufferedImage(thumbWidth, thumbHeight, image.TYPE_INT_RGB);
    result.getGraphics().drawImage(tempImage, 0, 0, null);
    
    var byteOut = new java.io.ByteArrayOutputStream();
    Packages.javax.imageio.ImageIO.write(result, "png", byteOut);
    image = new sun.misc.BASE64Encoder().encode(byteOut.toByteArray());
    return { big: atob(big),
             image: atob(("" + image).replace(/\r\n/g, "")) };
  },

  createServerThumb: function(wnd, thumbWidth, onFinish) {
    var data = FdThumbnail.createBase64Thumb(wnd);
    var request = "width=" + FdThumbnail.getWindowWidth(wnd) + 
                  "&thumbWidth=" + thumbWidth +
                  "&data=" + encodeURIComponent(data);
    FdUtils.ajax("http://userlogos.org/scale.php", "post", request,
                 function(response) {
                   onFinish({ big: atob(data), 
                              image: atob(response) });
                 });
  },

  createThumbnail: function(wnd, thumbWidth, onFinish) {
    var scaleMethod = FdPrefs.getCharPref("scaleMethod");
    if (scaleMethod == "server") {
      FdThumbnail.createServerThumb(wnd, thumbWidth, onFinish);
    } else if (scaleMethod == "java") {
      if (FdUtils.isJavaAvailable(function() {
                                    FdUtils.error("javaNotAvailable");
                                    onFinish();
                                  }))
      {
        onFinish(FdThumbnail.createJavaThumb(wnd, thumbWidth));
      }
    } else {
      onFinish(FdThumbnail.createCanvasThumb(wnd, thumbWidth));
    }
  }
}
