function $(id) {
  return document.getElementById(id);
}

var FdUtils = {
  clearElement: function(element) {
    for(var i = element.childNodes.length - 1; i >= 0; i--) {
      element.removeChild(element.childNodes[i]);
    }
  },

  merge: function(object1, object2) {
    var result = FdUtils.copy(null, object1);
    return FdUtils.copy(result, object2);
  },

  copy: function(target, source) {
    if (!target) target = new Object();

    for(var i in source) {
      if (source[i] == null) continue;
      switch(typeof source[i]) {
        case "string":
        case "number":
        case "boolean":
        case "function":
            target[i] = source[i];
            break;
        default:
            target[i] = FdUtils.copy(target[i], source[i]);
            break;
      }
    }
    return target;
  },

  confirm: function(message) {
    var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Components.interfaces.nsIPromptService);
    return prompts.confirm(window, FdBundle.getString("fastDial"), message);
  },

  confirmCheck: function(message, checkMessage, check) {
    var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Components.interfaces.nsIPromptService);
    return prompts.confirmCheck(window, FdBundle.getString("fastDial"), message, checkMessage, check);
  },

  getHiddenWindow: function() {
    return Components.classes["@mozilla.org/appshell/appShellService;1"]
           .getService(Components.interfaces.nsIAppShellService)
           .hiddenDOMWindow;
  },

  getBrowserWindow: function() {
    var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"]
             .getService(Components.interfaces.nsIWindowMediator);
    return wm.getMostRecentWindow("navigator:browser");
  },

  getFontList: function() {
    var enumerator = Components.classes["@mozilla.org/gfx/fontenumerator;1"]
                     .getService(Components.interfaces.nsIFontEnumerator);
    var count = {};
    return enumerator.EnumerateAllFonts(count);
  },

  isBlank: function(doc) {
    return doc && doc.location && doc.location.href == "about:blank";
  },

  forEachBlankPage: function(onPage) {
    for (var i = 0; i < gBrowser.browsers.length; i++) {
      var doc = gBrowser.browsers[i].contentDocument;
      if (FdUtils.isBlank(doc)) onPage(doc);
    }
  },

  isFirefox2: function() {
    return Components.classes["@mozilla.org/passwordmanager;1"];
  },

  getDocumentTab: function(doc) {
    var tabs = gBrowser.tabContainer.childNodes;
    for(var i = 0; i < tabs.length; i++) {
      if (tabs[i].linkedBrowser.contentDocument == doc) {
        return tabs[i];
      }
    }
  },

  whereToOpenLink: function(event) {
    if (!event) return "";

    if (event.button == 0) {
      return event.shiftKey ? (event.ctrlKey ? "tabshifted" : "window")
                            : (event.ctrlKey ? "tab" : "");
    } else if (event.button == 1) {
      return event.shiftKey ? "tab" : "tabshifted";
    }
  },

  // Works in both FF & SeaMonkey

  openUILink: function(url, where) {
    if (where instanceof Event) {
      where = FdUtils.whereToOpenLink(where);
    }
    switch(where) {
      case "tabshifted": return gBrowser.addTab(url);
      case "tab":        return gBrowser.selectedTab = gBrowser.addTab(url);
      case "window":     return open(url);
      default:           return loadURI(url);
    }
  },

  error: function(message, params) {
    try {
      message = FdBundle.getString(message, params);
    }
    catch(e) {}

    var fastDialError = FdBundle.getString("fastDialError");
    Components.utils.reportError(fastDialError + ": " + message);

    var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Components.interfaces.nsIPromptService);
    return prompts.alert(window, fastDialError, message);
  },

  isJavaAvailable: function(onFailure) {
    var failureTimeout;
    if (onFailure) {
      failureTimeout = setTimeout(onFailure, 0);
    }
    // If Java isn't installed, the following command causes silent javascript
    // interruption without throwing any exception
    var result = window.Packages &&
                 window.Packages.java != null;
    if (onFailure) {
      clearTimeout(failureTimeout);
      if (!result) onFailure();
    }
    return result;
  },

  ajax: function(url, method, content, onLoad) {
    var request = new XMLHttpRequest();
    request.onreadystatechange = function() {
      if (request.readyState == 4) {
        if (request.status == 200) {
          onLoad(request.responseText);
        } else {
          FdUtils.error("serverError", [request.status + " " + request.statusText]);
          onLoad();
        }
      }
    };
    request.open(method || "get", url, true);
    if (method == "post") {
      request.setRequestHeader("Content-Type", "application/x-www-form-urlencoded");
    }
    request.send(content);
  },

  toJSON: function(object, indent, level) {
    if (!level) level = 0;
    if (!indent) indent = " ";
    var json = "";

    for(var i in object) {
      var value = object[i];
      if (i[0] != "_" &&     // Field names beginning with "_" are ignored
          value != null &&
          typeof value != "function")
      {
        if (json) json += "," + indent;
        json += "'" + i + "': ";

        switch(typeof value) {
          case "number":
          case "boolean":
              json += value;
              break;
          case "string":
              json += "'" + value.replace(/([\\\'\n])/g, "\\$1") + "'";
              break;
          default:
              json += this.toJSON(value, indent, level + 1);
              break;
        }
      }
    }
    json = "{" + indent + json + indent + "}";
    return level ? json : "(" + json + ")";
  },

  getAbsolutePosition: function(element) {
    var position = {
      x: 0, y: 0,
      width: element.offsetWidth,
      height: element.offsetHeight,
    };

    do {
      position.x += element.offsetLeft;
      position.y += element.offsetTop;
    }
    while(element = element.offsetParent);

    return position;
  },

  getComputedStyle: function(element) {
    var doc = element.ownerDocument;
    var style = doc.defaultView.getComputedStyle(element, "");
    return {
      get: function(name) {
        return style.getPropertyValue(name);
      },
      getInt: function(name) {
        var value = style.getPropertyValue(name);
        return parseInt(value);
      }
    }
  },

  // Select item in menulist

  selectItem: function(menulistId, value, add) {
    var vl = String.toLowerCase(value);
    var popup = $(menulistId).menupopup;

    for(var i = 0; i < popup.childNodes.length; i++) {
      var l = String.toLowerCase(popup.childNodes[i].label);
      var v = String.toLowerCase(popup.childNodes[i].value);
      if (vl == l || vl == v) {
        popup.parentNode.selectedIndex = i;
        return;
      }
    }
    if (add && value) {
      var menuitem = popup.ownerDocument.createElement("menuitem");
      menuitem.setAttribute("label", value);
      menuitem.setAttribute("value", value);
      popup.parentNode.selectedItem = popup.appendChild(menuitem);
    }
    else popup.parentNode.selectedIndex = 0;
  },

  isMacOS: function() {
    var info = Components.classes["@mozilla.org/xre/app-info;1"]
               .getService(Components.interfaces.nsIXULRuntime);
    return info.OS == "Darwin";
  }
}

var FdBundle = {
  bundle: Components.classes["@mozilla.org/intl/stringbundle;1"]
          .getService(Components.interfaces.nsIStringBundleService)
          .createBundle("chrome://fastdial/locale/fastdial.properties"),

  getString: function(name, params) {
    try {
      return params ? FdBundle.bundle.formatStringFromName(name, params, params.length) 
                    : FdBundle.bundle.GetStringFromName(name);
    } catch(e) {
      return null;
    }
  }
}
